#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# test_mlc_analyzer.py
"""unittest cases for mlc_analyzer."""
#
# Copyright (c) 2020 Dan Cutright
# This file is part of DVHA-MLCA, released under a MIT license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVHA-MLCA


import unittest
from os.path import join
from mlca import mlc_analyzer, utilities
import pydicom
import json
from numpy.testing import assert_array_equal, assert_array_almost_equal

test_dir = "tests"
basedata_dir = join(test_dir, "testdata")
example_file_name = "rtplan.dcm"
example_file_path = join(basedata_dir, example_file_name)
mlc_borders_file = join(basedata_dir, "mlc_borders.json")
beam_summary_file = join(basedata_dir, "beam_summary.json")
plan_summary_file = join(basedata_dir, "plan_summary.json")


class TestMLCAnalzyer(unittest.TestCase):
    """Unit tests for Utilities."""

    def setUp(self):
        """Setup files and base data for utility testing."""
        self.data_path = "example_file_path"

        self.plan_ds = pydicom.read_file(example_file_path)

        with open(mlc_borders_file, "r") as fp:
            self.expected_mlc_borders = json.load(fp)

        with open(beam_summary_file, "r") as fp:
            self.expected_beam_summary = json.load(fp)

        with open(plan_summary_file, "r") as fp:
            self.expected_plan_summary = json.load(fp)

    def test_get_options(self):
        """Test get_options"""
        over_rides = {
            "max_field_size_x": 200,
            "max_field_size_y": 300,
            "complexity_weight_x": 0.5,
            "complexity_weight_y": 0.9,
        }
        self.assertEqual(over_rides, mlc_analyzer.get_options(over_rides))

        self.assertEqual(
            mlc_analyzer.DEFAULT_OPTIONS, mlc_analyzer.get_options({})
        )

        bad_over_rides = {"max_field_size_x": 200, "max_field_size_z": 300}
        expected = {
            key: value for key, value in mlc_analyzer.DEFAULT_OPTIONS.items()
        }
        expected["max_field_size_x"] = 200
        self.assertEqual(expected, mlc_analyzer.get_options(bad_over_rides))

    def test_control_point(self):
        """Test ControlPoint"""
        beam = self.plan_ds.BeamSequence[0]
        bld_seq = beam.BeamLimitingDeviceSequence
        leaf_boundaries = bld_seq[2].LeafPositionBoundaries

        cp = beam.ControlPointSequence[0]
        mlca_cp = mlc_analyzer.ControlPoint(cp, leaf_boundaries)

        # Compare jaw positions
        jaws = mlca_cp.jaws
        self.assertEqual(-78.1, jaws["x_min"])
        self.assertEqual(87.3, jaws["x_max"])
        self.assertEqual(-90.0, jaws["y_min"])
        self.assertEqual(100.0, jaws["y_max"])

        # Compare various aperture properties
        aperture = mlca_cp.aperture
        self.assertAlmostEqual(16172.0, aperture.area)
        self.assertAlmostEqual(770.7999999999998, aperture.length)
        self.assertAlmostEqual(27.931121691813, aperture.centroid.xy[0][0])
        self.assertAlmostEqual(14.66423447934702, aperture.centroid.xy[1][0])

        # check mlc borders compared to test data
        for key, value in mlca_cp.mlc_borders.items():
            assert_array_equal(self.expected_mlc_borders["0"][key], value)

        # CP comparison
        cp_2 = beam.ControlPointSequence[2]
        mlca_cp_2 = mlc_analyzer.ControlPoint(cp_2, leaf_boundaries)
        self.assertTrue(mlca_cp == mlca_cp)
        self.assertFalse(mlca_cp == mlca_cp_2)

    def test_beam(self):
        """Test Beam class"""
        expected_mu = 90.199996948242  # extracted from reference beam
        beam = mlc_analyzer.Beam(self.plan_ds.BeamSequence[0], expected_mu)

        # check beam summary to test data
        for key, value in beam.summary.items():
            assert_array_almost_equal(
                self.expected_beam_summary["0"][key], value
            )

        # check mlc borders compared to test data
        for key, value in beam.mlc_borders[0].items():
            assert_array_equal(self.expected_mlc_borders["0"][key], value)

        # compare two beams
        self.assertTrue(beam == beam)

        beam_wrong_mu = mlc_analyzer.Beam(self.plan_ds.BeamSequence[0], 1)
        self.assertFalse(beam == beam_wrong_mu)

        beam_2 = mlc_analyzer.Beam(self.plan_ds.BeamSequence[1], expected_mu)
        self.assertFalse(beam == beam_2)

        # test ignore zero MU CPs
        beam_no_zero = mlc_analyzer.Beam(
            self.plan_ds.BeamSequence[0], expected_mu, ignore_zero_mu_cp=True
        )
        exp_no_zero = [
            mu for mu in self.expected_beam_summary["0"]["cp_mu"] if mu > 0
        ]
        assert_array_equal(exp_no_zero, beam_no_zero.summary["cp_mu"])

    def test_fx_group(self):
        """Test of the FxGroup class"""
        beam_seq = self.plan_ds.BeamSequence
        fx_grp_seq = self.plan_ds.FractionGroupSequence
        fx_grp_1 = mlc_analyzer.FxGroup(fx_grp_seq[0], beam_seq)
        fx_grp_2 = mlc_analyzer.FxGroup(fx_grp_seq[1], beam_seq)
        self.assertFalse(fx_grp_1 == fx_grp_2)
        self.assertTrue(fx_grp_1 == fx_grp_1)

        self.assertEqual(9, fx_grp_1.beam_count)
        self.assertEqual(90.20, round(fx_grp_1.beam_mu[0], 2))
        self.assertEqual("rpo 200", fx_grp_1.beam_names[0])
        exp_cp_counts = [18, 10, 12, 14, 16, 16, 14, 20, 20]
        self.assertEqual(exp_cp_counts, fx_grp_1.cp_counts)
        self.assertEqual(776.10, round(fx_grp_1.fx_mu, 2))
        self.assertEqual("26", fx_grp_1.fxs)
        self.assertEqual(1.2742, round(fx_grp_1.younge_complexity_score, 4))

    def test_plan(self):
        """Test Plan class"""
        plan = mlc_analyzer.Plan(self.plan_ds)

        # test plan properties
        self.assertEqual("ANON11264", plan.patient_name)
        self.assertEqual("ANON11264", plan.patient_id)
        self.assertEqual("ADAC Pinnacle3", plan.tps)
        exp_cmp = [1.2741877055788196, 1.2897372167540244, 1.1271716301248724]
        assert_array_equal(exp_cmp, plan.younge_complexity_scores)

        # test plan equality comparison
        self.assertTrue(plan == plan)
        self.assertTrue(str(plan) == plan.__repr__())

        ds2 = pydicom.read_file(example_file_path)
        ds2.StudyInstanceUID = "test"
        plan_2 = mlc_analyzer.Plan(ds2)
        self.assertFalse(plan == plan_2)

        plan_2 = mlc_analyzer.Plan(self.plan_ds)
        plan_2.fx_group[0] = plan_2.fx_group[1]
        self.assertFalse(plan == plan_2)

        plan = mlc_analyzer.Plan(example_file_path)
        for i, fx_summary in enumerate(self.expected_plan_summary):
            for key, value in fx_summary.items():
                self.assertEqual(value, plan.summary[i][key])

    def test_plan_set(self):
        """Test PlanSet"""
        files = utilities.get_file_paths(test_dir)
        dcm_files = utilities.get_dicom_files(files, verbose=True)
        plan_set = mlc_analyzer.PlanSet(dcm_files, verbose=True)
        self.assertTrue(len(plan_set.summary_table) == 4)

        # Test _worker
        data = plan_set._worker(dcm_files[0])
        self.assertTrue(len(data) == 3)

    def test_plan_set_multiprocessing(self):
        """Test PlanSet with multiprocessing"""
        files = utilities.get_file_paths(test_dir)
        dcm_files = utilities.get_dicom_files(files, processes=2)
        plan_set = mlc_analyzer.PlanSet(dcm_files, verbose=True, processes=2)
        self.assertTrue(len(plan_set.summary_table) == 4)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# options.py
"""
Misc options for dvha-mlca
"""
# Copyright (c) 2016-2021 Dan Cutright
# This file is part of DVH Analytics MLC Analyzer, released under a BSD license
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVHA-MLCA


BEAM_MU_TOLERANCE = 0.001
CONTROL_POINT_MU_TOLERANCE = 0.00001
CONTROL_POINT_POS_TOLERANCE = 0.0001

DEFAULT_OPTIONS = {
    "max_field_size_x": 400.0,
    "max_field_size_y": 400.0,
    "complexity_weight_x": 1.0,
    "complexity_weight_y": 1.0,
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# mlc_analyzer.py
"""
Tools for analyzing beam and control point information from DICOM files
Hierarchy of classes: Plan -> FxGroup -> Beam -> ControlPoint
"""
# Copyright (c) 2016-2021 Dan Cutright
# This file is part of DVH Analytics MLC Analyzer, released under a BSD license
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVHA-MLCA

import pydicom
from pydicom.dataset import Dataset
import numpy as np
from shapely.geometry import Polygon
from shapely import speedups
from mlca.utilities import (
    flatten_list_of_lists as flatten,
    get_xy_path_lengths,
    run_multiprocessing,
)
from mlca.options import (
    BEAM_MU_TOLERANCE,
    CONTROL_POINT_MU_TOLERANCE,
    CONTROL_POINT_POS_TOLERANCE,
    DEFAULT_OPTIONS,
)
import warnings


COLUMNS = [
    "Patient Name",
    "Patient MRN",
    "Study Instance UID",
    "SOP Instance UID",
    "TPS",
    "Plan name",
    "# of Fx Group(s)",
    "Fx Group #",
    "Fractions",
    "Plan MUs",
    "Beam Count(s)",
    "Control Point(s)",
    "Complexity Score(s)",
    "File Name",
]


def get_options(over_rides):
    """Get MLC Analyzer options

    Parameters
    ----------
    over_rides : dict
        Over rides, keys may be 'max_field_size_x', 'max_field_size_y',
        'complexity_weight_x', or 'complexity_weight_y'

    Returns
    -------
    dict
        Options for field size and complexity weights. Default values are
        400 and 1.

    """
    options = {k: v for k, v in DEFAULT_OPTIONS.items()}
    for key, value in over_rides.items():
        if key in list(options):
            options[key] = value
    return options


# Enable shapely calculations using C, as opposed to the C++ default
if speedups.available:
    speedups.enable()


class PlanSet:
    """Parse DICOM-RT Plan files, analyze MLCs

    Parameters
    ----------
    file_paths : list
        A list of file paths to DICOM-RT Plan files
    verbose : bool, optional
        Set to true to print detailed information (ignored if
        multiprocessing enabled)
    processes : int
        Number of parallel processes allowed

    """

    def __init__(self, file_paths, verbose=False, processes=1, **kwargs):
        self.file_paths = file_paths
        self.verbose = verbose
        self.processes = processes
        self.kwargs = kwargs
        self.summary_table = [COLUMNS]

        if processes == 1:
            try:
                self._run()
            except KeyboardInterrupt:
                print("Plan analyzer halted!")
        else:
            data = run_multiprocessing(
                self._worker, self.file_paths, self.processes
            )
            for row in data:
                self.summary_table.extend(row)

    def _run(self):
        """Process files, accumulate data in self.summary_table"""
        plan_count = len(self.file_paths)
        for i, file_path in enumerate(self.file_paths):
            print("Analyzing (%s of %s): %s" % (i + 1, plan_count, file_path))
            try:
                plan = Plan(file_path, **self.kwargs)
                for fx_grp_row in plan.summary:
                    row = [fx_grp_row[key] for key in COLUMNS]
                    self.summary_table.append(row)

                if self.verbose:
                    print(plan, "\n")
            except Exception as e:
                print("Analysis failed\n%s\n" % e)

    def _worker(self, file_path):
        """Multiprocessing worker

        Parameters
        ----------
        file_path : str
            file path of a DICOM-RT Plan file

        Returns
        -------
        list
            Results from Plan.summary prepped for CSV output
        """
        data = []
        try:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                plan = Plan(file_path, **self.kwargs)
            for fx_grp_row in plan.summary:
                data.append([fx_grp_row[key] for key in COLUMNS])
        except Exception:
            pass
        return data


class Plan:
    """Collect plan information from an RT Plan DICOM file.
    Automatically parses fraction data with FxGroup class

    Parameters
    ----------
    rt_plan : str, Dataset
        file path of a DICOM RT Plan file or a pydicom Dataset

    """

    def __init__(self, rt_plan, **kwargs):

        if isinstance(rt_plan, Dataset):
            self.rt_plan_file = "Unknown"
        else:
            self.rt_plan_file = rt_plan
            rt_plan = pydicom.read_file(rt_plan)
        self.rt_plan = rt_plan

        self.options = get_options(kwargs)

        beam_seq = rt_plan.BeamSequence
        fx_grp_seq = rt_plan.FractionGroupSequence
        self.fx_group = [FxGroup(fx_grp, beam_seq) for fx_grp in fx_grp_seq]

        self.summary = [
            {
                "Patient Name": self.patient_name,
                "Patient MRN": self.patient_id,
                "Study Instance UID": self.study_instance_uid,
                "SOP Instance UID": self.sop_instance_uid,
                "TPS": self.tps,
                "Plan name": self.plan_name,
                "# of Fx Group(s)": str(len(self.fx_group)),
                "Fx Group #": str(f + 1),
                "Fractions": str(fx_grp.fxs),
                "Plan MUs": "%0.1f" % fx_grp.fx_mu,
                "Beam Count(s)": str(fx_grp.beam_count),
                "Control Point(s)": str(sum(fx_grp.cp_counts)),
                "Complexity Score(s)": "%0.3f"
                % self.younge_complexity_scores[f],
                "File Name": self.rt_plan_file,
            }
            for f, fx_grp in enumerate(self.fx_group)
        ]

    def __str__(self):
        summary = [
            "Patient Name:        %s" % self.patient_name,
            "Patient MRN:         %s" % self.patient_id,
            "Study Instance UID:  %s" % self.study_instance_uid,
            "SOP Instance UID:    %s" % self.sop_instance_uid,
            "TPS:                 %s" % self.tps,
            "Plan name:           %s" % self.plan_name,
            "# of Fx Group(s):    %s" % len(self.fx_group),
            "Plan MUs:            %s"
            % ", ".join(["%0.1f" % fx_grp.fx_mu for fx_grp in self.fx_group]),
            "Beam Count(s):       %s"
            % ", ".join([str(fx_grp.beam_count) for fx_grp in self.fx_group]),
            "Control Point(s):    %s"
            % ", ".join(
                [str(sum(fx_grp.cp_counts)) for fx_grp in self.fx_group]
            ),
            "Complexity Score(s): %s"
            % ", ".join(
                ["%0.3f" % cs for cs in self.younge_complexity_scores]
            ),
        ]
        return "\n".join(summary)

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        """Compare to Plan objects

        Parameters
        ----------
        other : Plan

        Returns
        -------
        bool
            True if all properties and FxGrp comparisons are true

        """

        attr = [
            "plan_name",
            "patient_name",
            "patient_id",
            "study_instance_uid",
            "sop_instance_uid",
            "tps",
        ]
        if not all([getattr(self, a) == getattr(other, a) for a in attr]):
            return False

        for i, fx_group in enumerate(self.fx_group):
            if not fx_group == other.fx_group[i]:
                return False
        return True

    @property
    def plan_name(self):
        """Get the plan name

        Returns
        -------
        str
            RTPlanLabel (300A,0002)
        """
        return str(getattr(self.rt_plan, "RTPlanLabel", ""))

    @property
    def patient_name(self):
        """Get the patient name

        Returns
        -------
        str
            PatientName (0010,0010)
        """
        return str(getattr(self.rt_plan, "PatientName", ""))

    @property
    def patient_id(self):
        """Get the patient id

        Returns
        -------
        str
            PatientID (0010,0020)
        """
        return str(getattr(self.rt_plan, "PatientID", ""))

    @property
    def study_instance_uid(self):
        """Get the patient study instance UID

        Returns
        -------
        str
            StudyInstanceUID (0020,000D)
        """
        return str(getattr(self.rt_plan, "StudyInstanceUID", ""))

    @property
    def sop_instance_uid(self):
        """Get the patient study instance UID

        Returns
        -------
        str
            SOPInstanceUID (0008,0018)
        """
        return str(getattr(self.rt_plan, "SOPInstanceUID", ""))

    @property
    def tps(self):
        """Get the treatment planning system

        Returns
        -------
        str
            Manufacturer (0008,0070) and ManufacturerModelName (0008,1090)
        """
        return "%s %s" % (
            getattr(self.rt_plan, "Manufacturer", ""),
            getattr(self.rt_plan, "ManufacturerModelName", ""),
        )

    @property
    def younge_complexity_scores(self):
        """Get the Younge complexity scores for each FxGroup

        Returns
        -------
        list
            FxGroup.younge_complexity_score for each fraction group
        """
        return [
            float(fx_grp.younge_complexity_score) for fx_grp in self.fx_group
        ]


class FxGroup:
    """Collect fraction group information from fraction group and beam
    sequences of a pydicom RT Plan dataset. Automatically parses beam data
    with Beam class

    Parameters
    ----------
    fx_grp_seq : Dataset
        element of FractionGroupSequence (300A,0070)
    plan_beam_sequences : BeamSequence
        BeamSequence (300A,00B0)

    """

    def __init__(self, fx_grp_seq, plan_beam_sequences, **kwargs):
        self.fxs = str(
            getattr(fx_grp_seq, "NumberOfFractionsPlanned", "UNKNOWN")
        )

        self.options = get_options(kwargs)

        meter_set = {}
        for ref_beam in fx_grp_seq.ReferencedBeamSequence:
            ref_beam_num = str(ref_beam.ReferencedBeamNumber)
            meter_set[ref_beam_num] = float(ref_beam.BeamMeterset)

        self.beam = []
        for beam in plan_beam_sequences:
            beam_num = str(beam.BeamNumber)
            if beam_num in meter_set:
                self.beam.append(Beam(beam, meter_set[beam_num]))
        self.update_missing_jaws()

    def __eq__(self, other):
        """Comparison of two FxGroup by comparing each beam

        Parameters
        ----------
        other : FxGroup
            Another FxGroup class object

        Returns
        -------
        bool
            True if all properties and Beam comparisons are true
        """

        attr = [
            "beam_count",
            "beam_names",
            "beam_mu",
            "fx_mu",
            "cp_counts",
            "younge_complexity_score",
        ]
        if not all([getattr(self, a) == getattr(other, a) for a in attr]):
            return False

        for i, beam in enumerate(self.beam):
            status_str = "beam %s\n\t%%s with other beam %s" % (
                self.beam_names[i],
                other.beam_names[i],
            )
            if not beam == other.beam[i]:
                print(status_str % "failed")
                return False
            else:
                print(status_str % "passed")
        return True

    @property
    def beam_count(self):
        """Get the number of beams

        Returns
        -------
        int
            Length of FxGroup.beams
        """
        return len(self.beam)

    @property
    def beam_names(self):
        """Get the beam names

        Returns
        -------
        list
            Beam.name for each beam
        """
        return [b.name for b in self.beam]

    @property
    def beam_mu(self):
        """Get the monitor units for each beam

        Returns
        -------
        list
            Beam.meter_set for each beam
        """
        return [b.meter_set for b in self.beam]

    @property
    def fx_mu(self):
        """Get the number of MU for for this fraction

        Returns
        -------
        float
            The sum of FxGroup.beam_mu
        """
        return np.sum(self.beam_mu)

    @property
    def cp_counts(self):
        """Get the number of control points for all beams

        Returns
        -------
        list
            Beam.cp_count for each beam
        """
        return [b.cp_count for b in self.beam]

    @property
    def younge_complexity_score(self):
        """Get the Younge complexity score this fraction

        Returns
        -------
        float
            The sum of Beam.younge_complexity_scores for all beams
        """
        return np.sum(
            np.array(
                [np.sum(beam.younge_complexity_scores) for beam in self.beam]
            )
        )

    def update_missing_jaws(self):
        """In plans with static jaws, jaw positions may
        not be found in each control point"""
        for i, beam in enumerate(self.beam):
            for j, cp in enumerate(beam.jaws):
                if (
                    cp["x_min"] == -self.options["max_field_size_x"] / 2
                    and cp["x_max"] == self.options["max_field_size_x"] / 2
                    and cp["y_min"] == -self.options["max_field_size_y"] / 2
                    and cp["y_max"] == self.options["max_field_size_y"] / 2
                ):
                    beam.jaws[j] = beam.jaws[0]


class Beam:
    """Collect beam information from a beam in a beam sequence of a pydicom
    RT Plan dataset. Automatically parses control point data with ControlPoint
    class

    Parameters
    ----------
    beam_dataset : Dataset
        element of a BeamSequence (300A,00B0)
    meter_set : int, float
        the monitor units for ``beam_dataset``
    ignore_zero_mu_cp : bool
        If True, skip over zero MU control points
        (e.g., as in Step-N-Shoot beams)

    """

    def __init__(
        self, beam_dataset, meter_set, ignore_zero_mu_cp=False, **kwargs
    ):
        self.beam_dataset = beam_dataset
        self.meter_set = meter_set
        self.ignore_zero_mu_cp = ignore_zero_mu_cp
        self.options = get_options(kwargs)

        self.control_point = [
            ControlPoint(cp, self.leaf_boundaries) for cp in self.cp_seq
        ]

        self.summary = {
            "cp": list(range(1, len(self.control_point) + 1)),
            "cum_mu_frac": [cp.cum_mu for cp in self.control_point],
            "cum_mu": self.cum_mu,
            "cp_mu": self.cp_mu,
            "gantry": self.gantry_angle,
            "collimator": self.collimator_angle,
            "couch": self.couch_angle,
            "jaw_x1": [j["x_min"] for j in self.jaws],
            "jaw_x2": [j["x_max"] for j in self.jaws],
            "jaw_y1": [j["y_min"] for j in self.jaws],
            "jaw_y2": [j["y_max"] for j in self.jaws],
            "area": self.area,
            "x_perim": self.perimeter_x,
            "y_perim": self.perimeter_y,
            "perim": self.perimeter,
            "cmp_score": self.younge_complexity_scores.tolist(),
        }

        for key in self.summary:
            if len(self.summary[key]) == 1:
                self.summary[key] = self.summary[key] * len(self.summary["cp"])

        if ignore_zero_mu_cp:
            non_zero_indices = [
                i
                for i, value in enumerate(self.summary["cp_mu"])
                if value != 0
            ]
            for key in list(self.summary):
                self.summary[key] = [
                    self.summary[key][i] for i in non_zero_indices
                ]

    def __eq__(self, other):
        """Compare ControlPoint classes in two beams

        Parameters
        ----------
        other : Beam
            Another Beam instance for comparison

        Returns
        -------
        bool
            True if same meter_set values and each paired ControlPoint elements
            evaluate as equal
        """
        if (self.meter_set - other.meter_set) > BEAM_MU_TOLERANCE:
            return False
        for i, cp in enumerate(self.control_point):
            if not cp == other.control_point[i]:
                print("cp %s failed" % i)
                return False
        return True

    @property
    def leaf_boundaries(self):
        """Get the leaf boundaries

        Returns
        -------
        DataSet
            LeafPositionBoundaries (300A,00BE)
        """
        for bld_seq in self.beam_dataset.BeamLimitingDeviceSequence:
            if hasattr(bld_seq, "LeafPositionBoundaries"):
                return bld_seq.LeafPositionBoundaries

    @property
    def aperture(self):
        """Get aperture shapely object for every control point

        Returns
        -------
        list
            ControlPoint.aperture for each control point

        """
        return [cp.aperture for cp in self.control_point]

    @property
    def name(self):
        """Get the Beam name

        Returns
        -------
        str
            In order of priority, BeamDescription (300A,00C3),
            BeamName (300A,00C2), or "Unknown"
        """
        return str(
            getattr(
                self.beam_dataset,
                "BeamDescription",
                getattr(self.beam_dataset, "BeamName", "Unknown"),
            )
        )

    @property
    def perimeter_x(self):
        """Get x-component of aperture perimeter for every control point

        Returns
        -------
        list
            ControlPoint.perimeter_x for each control point

        """
        return np.array([cp.perimeter_x for cp in self.control_point])

    @property
    def perimeter_y(self):
        """Get y-component of aperture perimeter for every control point

        Returns
        -------
        list
            ControlPoint.perimeter_y for each control point

        """
        return np.array([cp.perimeter_y for cp in self.control_point])

    @property
    def perimeter(self):
        """Get aperture perimeter for every control point

        Returns
        -------
        list
            ControlPoint.perimeter for each control point

        """
        return np.array([cp.perimeter for cp in self.control_point])

    @property
    def area(self):
        """Get aperture area for every control point

        Returns
        -------
        list
            ControlPoint.area for each control point

        """
        return [cp.area for cp in self.aperture]

    @property
    def cp_seq(self):
        """Get the control points

        Returns
        -------
        Dataset
            ControlPointSequence (300A,0111)
        """
        return self.beam_dataset.ControlPointSequence

    @property
    def cp_count(self):
        """Get the number of control points in this beam

        Returns
        -------
        int
            Length of ControlPointSequence (300A,0111)
        """
        return len(self.cp_seq)

    @property
    def jaws(self):
        """Jaw positions for each control point

        Returns
        -------
        list
            Each element is a ControlPoint.jaws dict
        """
        return [cp.jaws for cp in self.control_point]

    @property
    def cp_mu(self):
        """MU for each control point

        Returns
        -------
        np.ndarray
            Numpy array of control point MUs
        """
        return np.diff(np.array(self.cum_mu)).tolist() + [0]

    @property
    def mlc_borders(self):
        """Get the MLC border for each control point

        Returns
        -------
        list
            A list of ``ControlPoint.mlc_borders`` describing the boundaries of
            each leaf
        """
        return [cp.mlc_borders for cp in self.control_point]

    @property
    def gantry_angle(self):
        """Gantry angles for each control point

        Returns
        -------
        list
            A list of float values, defining gantry angles for each
            control point

        """
        return [
            float(cp.GantryAngle)
            for cp in self.cp_seq
            if hasattr(cp, "GantryAngle")
        ]

    @property
    def collimator_angle(self):
        """Collimator angles for each control point

        Returns
        -------
        list
            A list of float values, defining collimator angles for each
            control point

        """
        return [
            float(cp.BeamLimitingDeviceAngle)
            for cp in self.cp_seq
            if hasattr(cp, "BeamLimitingDeviceAngle")
        ]

    @property
    def couch_angle(self):
        """Couch angles for each control point

        Returns
        -------
        list
            A list of float values, defining couch angles for each
            control point

        """
        return [
            float(cp.PatientSupportAngle)
            for cp in self.cp_seq
            if hasattr(cp, "PatientSupportAngle")
        ]

    @property
    def cum_mu(self):
        """Cumulative monitor units for each control point

        Returns
        -------
        list
            A list of float values representing the cumulative MU
        """
        return [cp.cum_mu * self.meter_set for cp in self.control_point]

    @property
    def younge_complexity_scores(self):
        """Complexity score based on Younge et al

        Returns
        -------
        np.ndarray
            Younge complexity scores for each control point
        """
        # Complexity score based on:
        # Younge KC, Matuszak MM, Moran JM, McShan DL, Fraass BA,
        # Roberts DA. Penalization of aperture complexity in inversely planned
        # volumetric modulated arc therapy. Med Phys. 2012;39(11):7160–70.
        c1, c2 = (
            self.options["complexity_weight_x"],
            self.options["complexity_weight_y"],
        )
        if self.meter_set and self.meter_set > 0:
            return (
                np.divide(
                    np.multiply(
                        np.add(c1 * self.perimeter_x, c2 * self.perimeter_y),
                        self.cp_mu,
                    ),
                    self.area,
                )
                / self.meter_set
            )
        return 0


class ControlPoint:
    """Collect control point information from a ControlPointSequence in a beam
    dataset of a pydicom RT Plan dataset

    Parameters
    ----------
    cp_elem : DataElement
        element of a ControlPointSequence (300A,0111)
    leaf_boundaries : Dataset
        LeafPositionBoundaries (300A,00BE)

    """

    def __init__(self, cp_elem, leaf_boundaries, **kwargs):

        self.cp_elem = cp_elem
        self.leaf_boundaries = leaf_boundaries
        self.options = get_options(kwargs)

        self._set_leaf_jaw_type()

        self.path_lengths = get_xy_path_lengths(self.aperture)

    def _set_leaf_jaw_type(self):
        """Search for LeafJawPositions (300A,011C) assign
        ControlPoint.<RTBeamLimitingDeviceType (300A,00B8)>
        (i.e., x, y, asymx, asymy, mlcx, mlcy)
        Return type is ``np.ndarray`` or ``None`` for all"""

        if hasattr(self.cp_elem, "BeamLimitingDevicePositionSequence"):
            for (
                device_position_seq
            ) in self.cp_elem.BeamLimitingDevicePositionSequence:
                if hasattr(
                    device_position_seq, "RTBeamLimitingDeviceType"
                ) and hasattr(device_position_seq, "LeafJawPositions"):
                    leaf_jaw_type = str(
                        device_position_seq.RTBeamLimitingDeviceType
                    ).lower()
                    positions = np.array(
                        list(map(float, device_position_seq.LeafJawPositions))
                    )
                    mid_index = int(len(positions) / 2)
                    setattr(
                        self,
                        leaf_jaw_type,
                        [positions[:mid_index], positions[mid_index:]],
                    )

    @property
    def cum_mu(self):
        """Cumulative MU for this ControlPoint

        Returns
        -------
        float
            CumulativeMetersetWeight (300A,0134)

        """
        return float(self.cp_elem.CumulativeMetersetWeight)

    @property
    def mlc(self):
        """Get the MLC orientation

        Returns
        -------
        str, None
            Returns 'mlcx', 'mlcy' or ``None``
        """
        for leaf_type in ["mlcx", "mlcy"]:
            if hasattr(self, leaf_type):
                return getattr(self, leaf_type)

    @property
    def leaf_type(self):
        """Get the MLC orientation

        Returns
        -------
        str, None
            Returns 'mlcx', 'mlcy' or ``None``
        """
        for leaf_type in ["mlcx", "mlcy"]:
            if hasattr(self, leaf_type):
                return leaf_type

    @property
    def perimeter_x(self):
        """x-component of the aperture perimeter

        Returns
        -------
        float
            x-component of the Aperture perimeter
        """
        return self.path_lengths[0]

    @property
    def perimeter_y(self):
        """y-component of the aperture perimeter

        Returns
        -------
        float
            y-component of the Aperture perimeter
        """
        return self.path_lengths[1]

    @property
    def perimeter(self):
        """Perimeter of the aperture

        Returns
        -------
        float
            Aperture perimeter
        """
        return self.perimeter_x + self.perimeter_y

    def __eq__(self, other):
        """Compare to ControlPoint class objects

        Parameters
        ----------
        other : ControlPoint

        Returns
        -------
        bool
            True if cum_mu and each MLC position are with tolerance
        """

        if abs(self.cum_mu - other.cum_mu) > CONTROL_POINT_MU_TOLERANCE:
            return False
        for side in [0, 1]:
            for i, pos in enumerate(self.mlc[side]):
                if abs(pos - other.mlc[side][i]) > CONTROL_POINT_POS_TOLERANCE:
                    print(abs(pos - other.mlc[side][i]))
        return True

    @property
    def mlc_borders(self):
        """This function returns the boundaries of each MLC leaf for purposes
        of displaying a beam's eye view using bokeh's quad() glyph

        Returns
        -------
        dict
            the boundaries of each leaf within the control point with keys of
            'top', 'bottom', 'left', 'right

        """
        if self.mlc is not None:
            top = self.leaf_boundaries[0:-1] + self.leaf_boundaries[0:-1]
            top = [float(i) for i in top]
            bottom = self.leaf_boundaries[1::] + self.leaf_boundaries[1::]
            bottom = [float(i) for i in bottom]
            left = [-self.options["max_field_size_x"] / 2] * len(self.mlc[0])
            left.extend(self.mlc[1])
            right = self.mlc[0].tolist()
            right.extend(
                [self.options["max_field_size_x"] / 2] * len(self.mlc[1])
            )

            return {"top": top, "bottom": bottom, "left": left, "right": right}

    @property
    def aperture(self):
        """This function will return the outline of MLCs within jaws

        Returns
        -------
        Polygon
            a shapely object of the complete MLC aperture as one shape
            (including MLC overlap)

        """
        lb = self.leaf_boundaries
        mlc = self.mlc

        jaws = self.jaws
        jaw_points = [
            (jaws["x_min"], jaws["y_min"]),
            (jaws["x_min"], jaws["y_max"]),
            (jaws["x_max"], jaws["y_max"]),
            (jaws["x_max"], jaws["y_min"]),
        ]
        jaw_shapely = Polygon(jaw_points)

        if self.leaf_type == "mlcx":
            a = flatten(
                [[(m, lb[i]), (m, lb[i + 1])] for i, m in enumerate(mlc[0])]
            )
            b = flatten(
                [[(m, lb[i]), (m, lb[i + 1])] for i, m in enumerate(mlc[1])]
            )
        elif self.leaf_type == "mlcy":
            a = flatten(
                [[(lb[i], m), (lb[i + 1], m)] for i, m in enumerate(mlc[0])]
            )
            b = flatten(
                [[(lb[i], m), (lb[i + 1], m)] for i, m in enumerate(mlc[1])]
            )
        else:
            return jaw_shapely

        mlc_points = a + b[::-1]  # concatenate a and reverse(b)
        mlc_aperture = Polygon(mlc_points).buffer(0)

        # This function is very slow, since jaws are rectangular, perhaps
        # there's a more efficient method?
        aperture = mlc_aperture.intersection(jaw_shapely)

        return aperture

    @property
    def jaws(self):
        """Get the jaw positions of a control point

        Returns
        -------
        dict
            jaw positions (or max field size in lieu of a jaw). Keys are
            'x_min', 'x_max', 'y_min', 'y_max'

        """

        jaws = {}

        for dim in ["x", "y"]:
            half = self.options["max_field_size_%s" % dim] / 2.0
            values = getattr(self, "asym%s" % dim, [-half, half])
            jaws["%s_min" % dim] = float(min(values))
            jaws["%s_max" % dim] = float(max(values))

        return jaws



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# _version.py
"""Package initialization for DVHA-MLCA."""
# Copyright (c) 2021 Dan Cutright
# This file is part of dvha-mlca, released under a MIT license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVHA-MLCA

__author__ = "Dan Cutright"
__email__ = "dan.cutright@gmail.com"
__version__ = "0.2.3"
__release__ = "0.2.3"
__version_info__ = (0, 2, 3)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# utilities.py
"""
Utilities for DVHA-MLCA
"""
# Copyright (c) 2016-2021 Dan Cutright
# This file is part of DVH Analytics MLC Analyzer, released under a BSD license
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVHA-MLCA

import argparse
from datetime import datetime
from mlca._version import __version__
import numpy as np
import pydicom
from os import walk
from os.path import join
from mlca.options import DEFAULT_OPTIONS
from multiprocessing import Pool
from tqdm import tqdm
import warnings
import csv


def get_xy_path_lengths(shapely_object):
    """Get the x and y path lengths of a Shapely object

    Parameters
    ----------
    shapely_object : GeometryCollection, MultiPolygon, Polygon
        A shapely polygon-like object

    Returns
    -------
    list
        Perimeter lengths in the x and y directions

    """
    path = np.array([0.0, 0.0])
    if shapely_object.type == "GeometryCollection":
        for geometry in shapely_object.geoms:
            if geometry.type in {"MultiPolygon", "Polygon"}:
                path = np.add(path, get_xy_path_lengths(geometry))
    elif shapely_object.type == "MultiPolygon":
        for shape in shapely_object:
            path = np.add(path, get_xy_path_lengths(shape))
    elif shapely_object.type == "Polygon":
        x, y = np.array(shapely_object.exterior.xy[0]), np.array(
            shapely_object.exterior.xy[1]
        )
        path = np.array(
            [np.sum(np.abs(np.diff(x))), np.sum(np.abs(np.diff(y)))]
        )

    return path.tolist()


def flatten_list_of_lists(some_list, remove_duplicates=False, sort=False):
    """Convert a list of lists into one list of all values

    Parameters
    ----------
    some_list : list
        a list such that each element is a list
    remove_duplicates : bool, optional
        if True, return a unique list, otherwise keep duplicated values
    sort : bool, optional
        if True, sort the list

    Returns
    -------
    list
        A new list containing all values in ``some_list``

    """
    data = [item for sublist in some_list for item in sublist]

    if remove_duplicates:
        if sort:
            return list(set(data))
        else:
            ans = []
            for value in data:
                if value not in ans:
                    ans.append(value)
            return ans
    elif sort:
        return sorted(data)

    return data


def get_file_paths(init_dir):
    """Find all files in a directory and sub-directories

    Parameters
    ----------
    init_dir : str
        Top-level directory to search for files

    Returns
    -------
    list
        Absolute file paths

    """
    file_paths = []
    # iterate through files and all sub-directories
    for dirName, subdirList, fileList in walk(init_dir):
        for file_name in fileList:
            file_paths.append(join(dirName, file_name))
    return file_paths


def is_file_dicom(file_path, modality=None, verbose=False):
    """

    Parameters
    ----------
    file_path : str
        File path to potential DICOM file
    modality : str, optional
        Return False if file is not this Modality (0008,0060)
    verbose : bool, optional
        Print results to terminal

    Returns
    -------
    bool
        True if file_path points to a DICOM file, will return False if
        SOPClassUID (0008,0016) is not found

    """
    kwargs = {"stop_before_pixels": True, "force": True}
    try:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            ds = pydicom.read_file(file_path, **kwargs)
        # Assuming SOPClassUID is a required tag
        if modality is None and "SOPClassUID" in ds:
            if verbose:
                print("DICOM File Found: %s" % file_path)
            return True
        if ds.Modality.upper() == modality.upper():
            if verbose:
                print("DICOM %s File Found: %s" % (modality, file_path))
            return True
    except Exception as e:
        if verbose:
            print("Non-DICOM File Found: %s" % file_path)
            print(str(e))
    return False


def get_dicom_files(file_paths, modality=None, verbose=False, processes=1):
    """Find all DICOM-RT Plan files in a list of file paths

    Parameters
    ----------
    file_paths : list
        A list of file paths
    modality : str, optional
        Specify Modality (0008,0060)
    verbose : bool, optional
        Print results to terminal
    processes : int
        Number of processes for multiprocessing.

    Returns
    -------
    list
        Absolute file paths to DICOM-RT Plans

    """

    if processes == 1:
        return [f for f in file_paths if is_file_dicom(f, modality, verbose)]

    queue = [(f, modality, verbose) for f in file_paths]
    ans = run_multiprocessing(_get_dicom_files_worker, queue, processes)
    return [f for f in ans if f is not None]


def _get_dicom_files_worker(args):
    """Worker for get_dicom_files"""
    return args[0] if is_file_dicom(*args) else None


def run_multiprocessing(worker, queue, processes):
    """Parallel processing

    Parameters
    ----------
    worker : callable
        single parameter function to be called on each item in queue
    queue : iterable
        A list of arguments for worker
    processes : int
        Number of processes for multiprocessing.Pool

    Returns
    -------
    list
        List of returns from worker

    """
    progress_kwargs = {
        "total": len(queue),
        "bar_format": "{desc:<5.5}{percentage:3.0f}%|{bar:30}{r_bar}",
    }
    data = []
    with Pool(processes=processes) as pool:
        with tqdm(**progress_kwargs) as pbar:
            for item in pool.imap_unordered(worker, queue):
                data.append(item)
                pbar.update()
    return data


def create_cmd_parser():
    """Get an argument parser for mlca.main

    Returns
    -------
    argparse.ArgumentParser
        argument parser
    """
    cmd_parser = argparse.ArgumentParser(
        description="Command line DVHA MLC Analyzer"
    )
    cmd_parser.add_argument(
        "init_dir",
        nargs="?",
        help="Directory containing DICOM-RT Plan files",
        default=None,
    )
    cmd_parser.add_argument(
        "-of",
        "--output-file",
        dest="output_file",
        help="Output will be saved as dvha_mlca_<version>_results_"
        "<time-stamp>.csv by default.",
        default=None,
    )
    cmd_parser.add_argument(
        "-xw",
        "--x-weight",
        dest="complexity_weight_x",
        help="Complexity coefficient for x-dimension: default = %0.1f"
        % DEFAULT_OPTIONS["complexity_weight_x"],
        default=DEFAULT_OPTIONS["complexity_weight_x"],
    )
    cmd_parser.add_argument(
        "-yw",
        "--y-weight",
        dest="complexity_weight_y",
        help="Complexity coefficient for y-dimension: default = %0.1f"
        % DEFAULT_OPTIONS["complexity_weight_y"],
        default=DEFAULT_OPTIONS["complexity_weight_y"],
    )
    cmd_parser.add_argument(
        "-xs",
        "--x-max-field-size",
        dest="max_field_size_x",
        help="Maximum field size in the x-dimension: default = %0.1f (mm)"
        % DEFAULT_OPTIONS["max_field_size_x"],
        default=DEFAULT_OPTIONS["max_field_size_x"],
    )
    cmd_parser.add_argument(
        "-ys",
        "--y-max-field-size",
        dest="max_field_size_y",
        help="Maximum field size in the y-dimension: default = %0.1f (mm)"
        % DEFAULT_OPTIONS["max_field_size_y"],
        default=DEFAULT_OPTIONS["max_field_size_y"],
    )
    cmd_parser.add_argument(
        "-ver",
        "--version",
        dest="print_version",
        help="Print the DVHA-MLCA version",
        default=False,
        action="store_true",
    )
    cmd_parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        help="Print final results and plan summaries as they are analyzed",
        default=False,
        action="store_true",
    )
    cmd_parser.add_argument(
        "-n",
        "--processes",
        dest="processes",
        help="Enable multiprocessing, set number of parallel processes",
        default=1,
    )

    return cmd_parser


def get_default_output_filename():
    """Get the default output file name for mlca.main.process

    Returns
    -------
    str
        dvha_mlca_<version>_results_<timestamp>.csv
    """
    time_stamp = str(datetime.now().strftime("%Y-%m-%d_%H-%M-%S"))
    return "dvha_mlca_%s_results_%s.csv" % (
        __version__,
        time_stamp,
    )


def write_csv(file_path, rows, mode="w", newline=""):
    """Create csv.writer, call writerows(rows)

    Parameters
    ----------
    file_path : str
        path to file
    rows : list, iterable
        Items to be written to file_pointer (input for csv.writer.writerows)
    mode : str
        optional string that specifies the mode in which the file is opened
    newline : str
        controls how universal newlines mode works.
        It can be None, '', '\n', '\r', and '\r\n'
    """
    with open(file_path, mode, encoding="utf-8", newline=newline) as f:
        writer = csv.writer(
            f, delimiter=",", quotechar='"', quoting=csv.QUOTE_MINIMAL
        )
        writer.writerows(rows)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# main.py
"""
Collect DICOM-RT Plan Files, save summary to .csv
"""
# Copyright (c) 2016-2021 Dan Cutright
# This file is part of DVH Analytics MLC Analyzer, released under a BSD license
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVHA-MLCA

from mlca.mlc_analyzer import PlanSet
from mlca._version import __version__
from mlca.utilities import (
    get_file_paths,
    get_dicom_files,
    create_cmd_parser,
    get_default_output_filename,
    write_csv,
)


def process(
    init_dir=None,
    output_file=None,
    print_version=False,
    verbose=False,
    processes=1,
    **kwargs
):
    """Process command line args, call mlc_analyzer.PlanSet

    Parameters
    ----------
    init_dir : str
        Directory containing DICOM-RT Plan files
    output_file : str, optional
        Output will be saved as dvha_mlca_<version>_results_<time-stamp>.csv
        by default.
    print_version : bool, optional
        Print the DVHA-MLCA version
    verbose : bool, optional
        Print more detailed information as the script runs
    processes : int
        Number of processes used for multiprocessing
    """

    if print_version:
        print("DVHA MLC Analyzer: dvha-mlca v%s" % __version__)

    if init_dir is not None:

        # command line args are strings
        try:
            processes = int(float(processes))
        except Exception:
            processes = 1

        print("Directory: %s\n" "Begin file tree scan ..." % init_dir)
        file_paths = get_file_paths(init_dir)
        print(
            "File tree scan complete\n" "Searching for DICOM-RT Plan files ..."
        )
        dicom_plan_files = get_dicom_files(
            file_paths,
            modality="RTPLAN",
            verbose=verbose,
            processes=processes,
        )
        print("%s DICOM-RT Plan file(s) found" % len(dicom_plan_files))

        if not output_file:
            output_file = get_default_output_filename()

        kwargs["verbose"] = verbose
        kwargs["processes"] = processes
        print("Analyzing %s file(s) ..." % len(dicom_plan_files))
        plan_analyzer = PlanSet(dicom_plan_files, **kwargs)
        print("Analysis Complete")

        if kwargs["verbose"]:
            to_print = "\n".join(
                str(row) for row in plan_analyzer.summary_table
            )
            print(to_print.replace(",", "\t"))

        print("Printing summary to: %s" % output_file)
        write_csv(output_file, plan_analyzer.summary_table)

    elif not print_version:
        print("mlca: error: the following arguments are required: init_dir")


def main():
    """Parse command-line args, pass into process"""
    cmd_parser = create_cmd_parser()
    kwargs = vars(cmd_parser.parse_args())
    process(**kwargs)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# setup.py
"""
A setuptools setup file for DVH Analytics MLC Analyzer
"""
# Copyright (c) 2016-2021 Dan Cutright
# This file is part of DVH Analytics MLC Analyzer, released under a BSD license
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVHA-MLCA

from setuptools import setup, find_packages
from mlca._version import __version__


with open('requirements.txt', 'r') as doc:
    requires = [line.strip() for line in doc]

with open('README.rst', 'r') as doc:
    long_description = doc.read()

CLASSIFIERS = [
    "License :: OSI Approved :: MIT License",
    "Intended Audience :: End Users/Desktop",
    "Intended Audience :: Healthcare Industry",
    "Intended Audience :: Science/Research",
    "Natural Language :: English",
    "Development Status :: 4 - Beta",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Operating System :: OS Independent",
    "Topic :: Scientific/Engineering :: Medical Science Apps.",
    "Topic :: Scientific/Engineering :: Physics"]


setup(
    name='dvha-mlca',
    include_package_data=True,
    python_requires='>3.5',
    packages=find_packages(),
    version=__version__,
    description='Batch analyze DICOM-RT Plan files to calculate Complexity Scores',
    author='Dan Cutright',
    author_email='dan.cutright@gmail.com',
    url='https://github.com/cutright/DVHA-MLCA',
    download_url='https://github.com/cutright/DVHA-MLCA/archive/master.zip',
    license="MIT License",
    keywords=['radiation therapy', 'research', 'dicom', 'dicom-rt', 'analytics'],
    classifiers=CLASSIFIERS,
    install_requires=requires,
    entry_points={'console_scripts': ['mlca = mlca.main:main']},
    long_description=long_description,
    long_description_content_type="text/x-rst"
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# test_utilities.py
"""unittest cases for utilities."""
#
# Copyright (c) 2020 Dan Cutright
# This file is part of DVHA-MLCA, released under a MIT license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVHA-MLCA


import unittest
from os.path import join, basename
from os import unlink
from mlca import utilities
from mlca.options import DEFAULT_OPTIONS
from shapely.geometry import GeometryCollection, MultiPolygon, Polygon

test_dir = "tests"
basedata_dir = join(test_dir, "testdata")
example_file_name = "rtplan.dcm"
example_file_path = join(basedata_dir, example_file_name)


class TestUtilities(unittest.TestCase):
    """Unit tests for Utilities."""

    def setUp(self):
        """Setup files and base data for utility testing."""
        self.data_path = example_file_path

    def test_flatten_list_of_lists(self):
        """Test flatten_list_of_lists"""

        # keep all elements
        data = [[1, 10], [3, 5], [2, 3]]
        expected = [1, 10, 3, 5, 2, 3]
        calculated = utilities.flatten_list_of_lists(data)
        self.assertEqual(expected, calculated)

        # remove duplicates
        expected = [1, 10, 3, 5, 2]
        calculated = utilities.flatten_list_of_lists(
            data, remove_duplicates=True
        )
        self.assertEqual(expected, calculated)

        # sorted
        expected = [1, 2, 3, 3, 5, 10]
        calculated = utilities.flatten_list_of_lists(data, sort=True)
        self.assertEqual(expected, calculated)

        # sorted and removed duplicates
        expected = [1, 2, 3, 5, 10]
        calculated = utilities.flatten_list_of_lists(
            data, remove_duplicates=True, sort=True
        )
        self.assertEqual(expected, calculated)

    def test_get_dicom_files(self):
        """Test get_dicom_rt_plan_files"""
        files = utilities.get_file_paths(test_dir)
        dcm_files = utilities.get_dicom_files(files, verbose=True)
        self.assertTrue(len(dcm_files) == 1)
        self.assertEqual(basename(self.data_path), basename(dcm_files[0]))

        dcm_files = utilities.get_dicom_files(files, "RTPLAN", True)
        self.assertTrue(len(dcm_files) == 1)
        self.assertEqual(basename(self.data_path), basename(dcm_files[0]))

    def test_get_dicom_files_worker(self):
        """Test get_dicom_rt_plan_files"""
        files = utilities.get_file_paths(test_dir)
        args = (files[0], "TEST", False)
        ans = utilities._get_dicom_files_worker(args)
        self.assertIsNone(ans)

    def test_get_xy_path_lengths(self):
        """Test get_xy_path_lengths"""
        # paths lengths -> x = 2, y = 4
        polygon = Polygon([[0, 0], [1, 0], [1, 2], [0, 2], [0, 0]])

        # paths lengths -> x = 4, y = 8
        multi_polygon = MultiPolygon([polygon, polygon])

        # paths lengths -> x = 6, y = 12
        collection = GeometryCollection([multi_polygon, polygon])

        path_lengths = utilities.get_xy_path_lengths(collection)
        self.assertEqual([6.0, 12.0], path_lengths)

    def test_create_cmd_parser(self):
        """Test create_cmd_parser"""
        cmd_parser = utilities.create_cmd_parser()
        kwargs = vars(cmd_parser.parse_args([]))
        keys = sorted(list(kwargs))
        exp = sorted(
            [
                "init_dir",
                "output_file",
                "complexity_weight_x",
                "complexity_weight_y",
                "max_field_size_x",
                "max_field_size_y",
                "print_version",
                "verbose",
                "processes",
            ]
        )
        self.assertEqual(keys, exp)
        for key in ["complexity_weight_", "max_field_size_"]:
            for dim in ["x", "y"]:
                self.assertEqual(kwargs[key + dim], DEFAULT_OPTIONS[key + dim])

    def test_get_default_output_filename(self):
        """test get_default_output_filename"""
        file_name = utilities.get_default_output_filename()
        self.assertTrue(isinstance(file_name, str))
        self.assertEqual(file_name.split(".")[-1], "csv")

    def test_write_csv(self):
        """test write_csv"""
        file_path = "test_write_csv"
        utilities.write_csv("test_write_csv", [["one", "two"], [1, 2]])
        with open(file_path, "r") as f:
            data = f.readlines()

        self.assertEqual(data[0].strip(), "one,two")
        self.assertEqual(data[1].strip(), "1,2")

        try:
            unlink(file_path)
        except Exception:
            pass
